# Write shell commands in this file
# The command name in the shell is routed to cmd_+command name in this class
# All commands take one argument which is a list of space-split arguments
# preserving quoted substrings.
# E.g., if you type:
# hello abc "abd askdjf"
# we will call cmd_hello(['abc','abd askdjf'])
# Methods should return a true value or false to abort the shell and close connection
import base64
import crypto
import json
import keymanager

def encode_file(ftype, data):
    data = base64.b64encode(data)
    return json.dumps({'ftype': ftype, 'data': data})

def decode_file(data):
    d = json.loads(data)
    d['data'] = base64.b64decode(d['data'])
    return d

class CommandHandler(object):
    def __init__(self, connection, user_id=None, home_id=None, pubKey=None, priKey=None):
        self.conn = connection
        self.user_id = user_id
        self.user_name = ''
        self.home_id = home_id
        self.pubKey = pubKey
        self.priKey = priKey
        self.working_directory_id = home_id
        self.working_directory_name = ''
        self.working_directory_conts = []
        self.directory_ids = {}
        self.directory_hist = []
        self.km = keymanager.KeyManager('keys.txt', 'users.txt')

        if self.working_directory_id:
            result = self.read_file(self.working_directory_id)
            if not result:
                return
            self.user_name = result['name']
            self.working_directory_name = '/' + result['name']
            self.working_directory_conts = json.loads(result['data']['data'])

            files = json.loads(result['data']['data'])
            for f in files:
                r_f = self.read_file(f)
                if r_f:
                    self.directory_ids[r_f['name']] = f

    def cmd_exit(self, data):
        return False

    def cmd_update(self, data):
        if len(data) != 0:
            print 'Usage: update'
            return True
        result = self.conn.cmd({'method': 'update'})
        for k in result['user_keys'].keys():
            e = self.km.get_user_key(k)
            if e == None:
                self.km.set_user_key(k, result['user_keys'][k])
            elif e != result['user_keys'][k]:
                raise RuntimeError('Server issues: keys have been changed')
        files = json.loads(result['pending_list'])
        keys = json.loads(result['pending_keys'])
        
        for i in range(len(files)):
            file_id = files[i]
            file_key = crypto.rsa_decrypt(self.priKey, base64.b64decode(keys[i]))
            file_key = json.loads(file_key)
            file_key['key'] = base64.b64decode(file_key['key'])
            file_key['iv'] = base64.b64decode(file_key['iv'])
            self.km.set_key(file_id, file_key)

            # Update the working directory to contain the file
            if self.working_directory_id is not None:
                self.working_directory_conts += [file_id]
                self.update_file(self.working_directory_id, 'dir', json.dumps(self.working_directory_conts))
                r_f = self.read_file(file_id)                
                self.directory_ids[r_f['name']] = file_id

        return True

    def cmd_register(self, data):
        if len(data) != 1:
            print 'Usage: register username'
            return True
        if self.user_id is not None:
            print 'Already registered'
            return True
        # Generate keys
        self.user_name = username = data[0]
        key = crypto.rsa_generate()
        f = open('pubKey.pem', 'w')
        f.write(key['pubKey'])
        self.pubKey = key['pubKey']
        f.close()
        f = open('priKey.pem', 'w')
        f.write(key['priKey'])
        self.priKey = key['priKey']

        # Register on the server
        result = self.conn.cmd({'method': 'register',
                                'username': username,
                                'pubkey': key['pubKey']})
        if result == False:
            return False

        self.user_id = result['id']
        self.conn.set_key(self.user_id, self.priKey)
        for k in result['user_keys'].keys():
            self.km.set_user_key(k, result['user_keys'][k])

        # Create home directory for the user
        home_dir_result = self.mkdir_home([username])
        home_dir_id = home_dir_result['id']
        self.home_id = home_dir_id
        self.working_directory_id = home_dir_id
        self.working_directory_name = "/" + username
        self.working_directory_conts = []

        f = open('idhome.txt', 'w')
        f.write(str(result['id']) + '\n' + str(home_dir_id))
        f.close()
        return True

    def mkdir_home(self, data):
        dirname = data[0]
        acl = {self.user_name : 'w'}
        result = self.create_file(dirname, 'dir', json.dumps([]), acl)
        return result

    def cmd_mkdir(self, data):
        self.cmd_update([])
        if len(data) != 1:
            print 'Usage: mkdir name'
            return True
        dirname = data[0]
        acl = {self.user_name : 'w'}
        users = raw_input("Share with: ")
        if users != '':
            for user in users.split(' '):
                counter = 0
                while(True):

                    perm = raw_input("Permissions for %s (r/w)" % user + (" (you have failed %d times): " \
                                     % counter if counter > 0 else ": "))
                    if perm in ['r', 'w', 'rw']:
                        break
                    counter += 1
                acl[user] = perm
        result = self.create_file(dirname, 'dir', json.dumps([]), acl)
        return result

    def cmd_cd(self, data):
        self.cmd_update([])
        if len(data) != 1:
            print 'Usage: cd DIRECTORY'
            return True
        if data[0] == "..":
            if len(self.directory_hist) == 0:
                print "Already in root directory"
                return True
            else:
                self.working_directory_id = self.directory_hist[-1]
                self.directory_hist.pop()
                text = self.working_directory_name.split('/')
                self.working_directory_name = '/'.join(text[0:-2])
        elif data[0] not in self.directory_ids:
            print "Directory not in current working directory"
            return True
        else:
            self.directory_hist.append(self.working_directory_id)
            self.working_directory_id = self.directory_ids[data[0]] 
        
        result = self.read_file(self.working_directory_id)
        
        if not result or len(result) == 0:
            print "Directory doesn't exist"
            return True

        self.working_directory_name += "/" + result['name']
        self.working_directory_conts = json.loads(result['data']['data'])

        files = json.loads(result['data']['data'])
        self.directory_ids = {}
        for f in files:
            r_f = self.read_file(f)
            if r_f:
                self.directory_ids[r_f['name']] = f
        return True

    def cmd_ls(self, data):
        result = self.read_file(self.working_directory_id)
        
        if not result or len(result) == 0:
            print "Directory doesn't exist"
            return True

        self.working_directory_conts = json.loads(result['data']['data'])

        files = json.loads(result['data']['data'])
        self.directory_ids = {}
        for f in files:
            r_f = self.read_file(f)
            if r_f:
                self.directory_ids[r_f['name']] = f

        self.cmd_update([])
        if len(data) > 0:
            print 'Usage: ls'
            return True
        for key in self.directory_ids.keys():
            print key
        return True

    def cmd_upload(self, data):
        self.cmd_update([])
        if len(data) < 1 or len(data) > 2:
            print 'Usage: upload PATH FILENAME(optional)'
            return True
        path = data[0]
        name = path.split('/')[-1] if len(data) == 1 else data[1]
        if name not in self.directory_ids.keys():
            acl = {self.user_name : 'w'}
            users = raw_input("Share with: ")
            if users != '':
                for user in users.split(' '):
                    counter = 0
                    while(True):

                        perm = raw_input("Permissions for %s (r/w)" % user + (" (you have failed %d times): " \
                                         % counter if counter > 0 else ": "))
                        if perm in ['r', 'w', 'rw']:
                            break
                        counter += 1
                    acl[user] = perm
            f = open(path, 'r')
            result = self.create_file(name, 'file', f.read(), acl)
            if not result:
                print "File creation failed"
                return
            else:
                print "Created file id %d" % result['id']
        else:
            f = open(path, 'r')
            self.update_file(self.directory_ids[name], 'file', f.read())
        return True

    def cmd_mv(self, data):
        self.cmd_update([])
        if len(data) != 2:
            print "Usage: mv FILENAME NEWNAME"
            return True
        name = data[0]
        new_name = data[1]
        if name not in self.directory_ids:
            print "%s not in current directory" % name
            return True
        self.rename_file(self.directory_ids[name], name, new_name)
        return True

    def cmd_download(self, data):
        self.cmd_update([])
        if len(data) != 1:
            print "Usage: download FILENAME"
            return True
        name = data[0]
        d = self.read_file(self.directory_ids[name])
        if len(d) == 0:
            print "File does not exist"
            return True
        f = open(name, 'w')
        f.write(d['data']['data'])
        f.close()
        return True

    def cmd_rm(self, data):
        self.cmd_update([])
        if len(data) != 1:
            print "Usage: rm FILENAME"
            return True
        name = data[0]
        if name not in self.directory_ids.keys():
            print "File not in local directory"
            return True
        result = self.conn.cmd({
            'method': 'delete_file',
            'id': self.directory_ids[name]
        })
        if not result:
            print "Rm failed entirely"
            return False
        if result['id'] == -1 or result['id'] == -2:
            r = self.read_file(self.working_directory_id)                
            self.working_directory_conts.remove(self.directory_ids[name])
            self.update_file(self.working_directory_id, 'dir', json.dumps(self.working_directory_conts))
            del self.directory_ids[name]
        if result['id'] == -3:
            if name in self.directory_ids.keys():
                del self.directory_ids[name]
        return True

    def create_file(self, name, ftype, data, acl):
        fdata = encode_file(ftype, data)
        keys = crypto.aes_generate()
        ekeys = dict(keys)
        ekeys['key'] = base64.b64encode(ekeys['key'])
        ekeys['iv'] = base64.b64encode(ekeys['iv'])

        sks = {}
        for k in acl.keys():
            if k != self.user_name:
                key = self.km.get_user_key(k)
                if key:
                    d = json.dumps(ekeys)
                    sks[k] = base64.b64encode(crypto.rsa_encrypt(key, d))

        acl = json.dumps(acl)
        acl_sig = base64.b64encode(crypto.rsa_sign(self.priKey, acl))
        orig_name = name
        name = crypto.aes_encrypt(keys['key'], keys['iv'], name)
        name_sig = crypto.rsa_sign(self.priKey, name)
        fdata = crypto.aes_encrypt(keys['key'], keys['iv'], fdata)
        data_sig = crypto.rsa_sign(self.priKey, fdata)

        name_sig = base64.b64encode(name_sig)
        name = base64.b64encode(name)
        data_sig = base64.b64encode(data_sig)
        fdata = base64.b64encode(fdata)


        result = self.conn.cmd({
            'method': 'create_file',
            'owner_id': self.user_id,
            'name': name,
            'name_signature': name_sig,
            'acl': acl,
            'acl_signature': acl_sig,
            'data': fdata,
            'data_signature': data_sig,
            'secret_keys': json.dumps(sks)
        })
        if not result:
            return False
        self.km.set_key(result['id'], keys)

        # Update the working directory to contain the file
        if self.working_directory_id is not None:
            self.working_directory_conts += [result['id']]
            self.update_file(self.working_directory_id, 'dir', json.dumps(self.working_directory_conts))
            self.directory_ids[orig_name] = result['id']

        return result

    def update_file(self, file_id, ftype, data):
        fdata = encode_file(ftype, data)
        keys = self.km.get_key(file_id)
        fdata = crypto.aes_encrypt(keys['key'], keys['iv'], fdata)
        data_sig = base64.b64encode(crypto.rsa_sign(self.priKey, fdata))
        fdata = base64.b64encode(fdata) 

        result = self.conn.cmd({
            'method': 'update_file',
            'id': file_id,
            'data': fdata,
            'data_signature': data_sig
        })

        return result

    def rename_file(self, file_id, ori_name, new_name):
        keys = self.km.get_key(file_id)
        name = crypto.aes_encrypt(keys['key'], keys['iv'], new_name)
        name_sig = base64.b64encode(crypto.rsa_sign(self.priKey, name))
        name = base64.b64encode(name)
        result = self.conn.cmd({
            'method': 'rename_file',
            'id': file_id,
            'name': name,
            'name_signature': name_sig
        })
        if not result:
            return False
        if len(result.keys()) != 0:
            self.directory_ids[new_name] = file_id
            del self.directory_ids[ori_name]
        return result

    def read_file(self, file_id):
        keys = self.km.get_key(file_id)
        if keys is None:
            return False

        data = self.conn.cmd({'method': 'read_file', 'id': file_id})
        if 'id' in data.keys() and data['id'] == -1:
            r = self.read_file(self.working_directory_id)                
            self.working_directory_conts.remove(file_id)
            self.update_file(self.working_directory_id, 'dir', json.dumps(self.working_directory_conts))
            for key in self.directory_ids.keys():
                if self.directory_ids[key] == file_id:
                    del self.directory_ids[key]
            #print "WARNING: file does not exist - removing from directory"
            return {}
        writers = []
        for k in json.loads(data['acl']).keys():
            if json.loads(data['acl'])[k] == 'w' or json.loads(data['acl'])[k] == 'rw':
                writers.append(k)

        # verify signatures
        if not crypto.rsa_verify(self.km.get_user_key(data['owner']), base64.b64decode(data['acl_signature']), data['acl']) or \
            all([not crypto.rsa_verify(self.km.get_user_key(w), base64.b64decode(data['name_signature']), base64.b64decode(data['name'])) for w in writers]) or \
            all([not crypto.rsa_verify(self.km.get_user_key(w), base64.b64decode(data['data_signature']), base64.b64decode(data['data'])) for w in writers]): \
            raise RuntimeError('Server issues: signatures not valid')

        name_dec = crypto.aes_decrypt(keys['key'], keys['iv'], base64.b64decode(data['name']))
        data_dec = crypto.aes_decrypt(keys['key'], keys['iv'], base64.b64decode(data['data']))
        data_dec = decode_file(data_dec)

        return {'owner': data['owner'], 'name': name_dec,
                'acl': data['acl'], 'data': data_dec}
