#!/usr/bin/env python

import os
from sqlalchemy import create_engine, ForeignKey
from sqlalchemy import Column, Date, Integer, String
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker

FILE_DB = 'files.db'
DEBUG_DB = False
Base = declarative_base()

class User(Base):
    __tablename__ = "user"

    id = Column(Integer, primary_key=True)
    username = Column(String, unique=True)
    pubkey = Column(String)
    pending_list = Column(String)
    pending_keys = Column(String)

    def __init__(self, username, pubkey, pending_list, pending_keys):
        self.username = username
        self.pubkey = pubkey
        self.pending_list = pending_list
        self.pending_keys = pending_keys

class File(Base):
    __tablename__ = "file"

    id = Column(Integer, primary_key=True)
    owner_id = Column(ForeignKey('user.id'))
    name = Column(String)
    name_signature = Column(String)
    acl = Column(String)
    acl_signature = Column(String)
    data = Column(String)
    data_signature = Column(String)
    

    def __init__(self, owner_id, name, name_signature, acl, acl_signature, data, data_signature):
        self.owner_id = owner_id
        self.name = name
        self.name_signature = name_signature
        self.acl = acl
        self.acl_signature = acl_signature
        self.data = data
        self.data_signature = data_signature

def get_engine():
    return create_engine('sqlite:///' + FILE_DB, echo=DEBUG_DB) 

def get_db():
    engine = get_engine()
    Session = sessionmaker(bind=engine)
    return Session()

if __name__ == '__main__':
    if raw_input('Recreate DB? (y/n) ') == 'y':
        os.unlink(FILE_DB)
        Base.metadata.create_all(get_engine())
