#!/usr/bin/env python

from handle import RequestHandler
from db import get_db, User
import crypto
import json
import base64
import SocketServer
import sys

HOST = 'localhost'
PORT = 26858

def verify_signed(data, user_id, signature, db):
    u = db.query(User).get(user_id)
    verify = crypto.rsa_verify(u.pubkey, signature, data)
    return verify


class TCPConnectionHandler(SocketServer.StreamRequestHandler):
    def respond(self, obj):
        if type(obj) != dict:
            obj = {'response': obj}
        self.wfile.write(json.dumps(obj) + '\n')

    # db to pass to signature verifier
    def parse_req(self, data, db):
        uid = -1
        try:
            dsplit = data.split(' ')
            signature = dsplit[-1].strip()
            data = ' '.join(dsplit[0:-1])
            if len(signature) > 0:
                sig = base64.b64decode(signature)
                uid, sign = json.loads(sig)
                sign = base64.b64decode(sign)
                signed = verify_signed(data, uid, sign, db)
            else:
                signed = False
                data = ' '.join(dsplit)
            data = ' '.join(dsplit[0:-1])
            decoded = json.loads(data)
        except ValueError:
            raise RuntimeError('')
        if type(decoded) != dict:
            raise RuntimeError('Request is not a dictionary')
        if 'method' not in decoded.keys():
            raise RuntimeError('No method call')
        return (signed, decoded, uid)

    def handle(self):
        req_handler = RequestHandler()
        db = get_db()
        while True:
            data = self.rfile.readline()
            if data is not None:
                try:
                    signed, req, uid = self.parse_req(data, db)
                    if not signed and req['method'] != 'register':
                        raise RuntimeError('Invalid signature')
                    if 'req_' + req['method'] in dir(req_handler):
                        req['uid'] = uid
                        response = getattr(req_handler, 'req_' + req['method'])(req)
                        self.respond(response)
                    else:
                        raise RuntimeError('No method found')
                except RuntimeError, e:
                    self.respond({'error': e.message})
                    self.request.close()
                    break
        req_handler.cleanup()


class Server(SocketServer.ThreadingMixIn, SocketServer.TCPServer):
    daemon_threads = True
    allow_reuse_address = True

    def __init__(self, server_address, RequestHandlerClass):
        SocketServer.TCPServer.__init__(self, server_address, RequestHandlerClass)


if __name__ == '__main__':
    server = Server((HOST, PORT), TCPConnectionHandler)
    try:
        server.serve_forever()
    except KeyboardInterrupt:
        sys.exit(0)
